/*************************************************************************
 * The contents of this file are subject to the MYRICOM SNIFFER10G
 * LICENSE (the "License"); User may not use this file except in
 * compliance with the License.  The full text of the License can found
 * in LICENSE.TXT
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied.  See
 * the License for the specific language governing rights and
 * limitations under the License.
 *
 * Copyright 2015 by CSP, Inc.  All rights reserved.
 ***********************************************************************/

/* A simplified version of the private mal_ headers to compile
 * SNF examples in the install directory.
 */
#ifdef _WIN32

#include <windows.h>
#include <process.h>

typedef unsigned long long mal_cycles_t;

static inline mal_cycles_t mal_get_cycles(void)
{
  LARGE_INTEGER li;

  QueryPerformanceCounter(&li);
  return li.QuadPart;
}

static inline mal_cycles_t mal_cycles_per_second(void)
{
  LARGE_INTEGER li;

  QueryPerformanceFrequency(&li);
  return li.QuadPart;
}

#else /* _WIN32 */

#include <inttypes.h>
#include <pthread.h>
#include <sys/time.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>

typedef uint64_t mal_cycles_t;

/* Linux on x86 */
static inline mal_cycles_t mal_get_cycles(void)
{
  unsigned l,h;
  __asm__ __volatile__("rdtsc": "=a" (l), "=d" (h));
  return l + (((mal_cycles_t)h) << 32);
}

mal_cycles_t mal_cycles_per_second(void);

#define MAL_MUTEX_T pthread_mutex_t
#define MAL_MUTEX_INIT(lock) pthread_mutex_init(lock,NULL)
#define MAL_MUTEX_DESTROY(lock) pthread_mutex_destroy(lock)
#define MAL_MUTEX_LOCK(lock) pthread_mutex_lock(lock)
#define MAL_MUTEX_UNLOCK(lock) pthread_mutex_unlock(lock)

#define MAL_THREAD_T pthread_t
#define MAL_THREAD_RETURN_T void*
#define MAL_THREAD_CREATE(threadp,fctn,arg,name) \
pthread_create(threadp, NULL, fctn, arg)
#define MAL_THREAD_CREATE_BIND(threadp,fctn,arg,name,cpu) \
pthread_create(threadp, NULL, fctn, arg)
#define MAL_THREAD_JOIN(threadp) pthread_join(*(threadp), NULL)
#define MAL_THREAD_EXIT_WITH_NOTHING(threadp) pthread_exit(0); return 0
#define MAL_THREAD_SELF() pthread_self()
#define MAL_THREAD_EQUAL(t1, t2) pthread_equal(t1, t2)
#define MAL_THREAD_CANCEL(threadp) pthread_cancel(*threadp)
#define MAL_THREAD_INIT(threadp, name) do { /* nothing */ } while (0)

#define MAL_PRINT(a) printf a

#endif

void mal_cycles_counter_init(void);
