/*************************************************************************
 * The contents of this file are subject to the MYRICOM MYRINET          *
 * EXPRESS (MX) NETWORKING SOFTWARE AND DOCUMENTATION LICENSE (the       *
 * "License"); User may not use this file except in compliance with the  *
 * License.  The full text of the License can found in LICENSE.TXT       *
 *                                                                       *
 * Software distributed under the License is distributed on an "AS IS"   *
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied.  See  *
 * the License for the specific language governing rights and            *
 * limitations under the License.                                        *
 *                                                                       *
 * Copyright 2003 - 2010 by Myricom, Inc.  All rights reserved.          *
 *************************************************************************/

#ifndef _counters_h_
#define _counters_h_

#include "mal_auto_config.h"


#if defined(COUNTERS_VAL)

#define COUNTERS_DEF(id) _COUNTERS_DEF(id)
#define _COUNTERS_DEF(id) struct mcp_counters_##id
#define COUNTER(field, cclass, label) uint32_t field;

#elif defined(COUNTERS_STR)

#define COUNTERS_DEF(id) _COUNTERS_DEF(id)
#define _COUNTERS_DEF(id) const char *mcp_counters_##id##_str[] = 
#define COUNTER(field, cclass, label) label,

#elif defined(COUNTERS_CLASS)

#define COUNTERS_DEF(id) _COUNTERS_DEF(id)
#define _COUNTERS_DEF(id) const uint32_t mcp_counters_##id##_class[] = 
#define COUNTER(field, cclass, label) MCP_COUNTER_CLASS_##cclass,

#else /* MCP */

#define COUNTERS_DEF(id) struct mcp_counters
#define COUNTER(field, cclass, label) uint32_t field;

#endif /* MCP */

#define MCP_COUNTER_CLASS_NORMAL 0
#define MCP_COUNTER_CLASS_RAW    1
#define MCP_COUNTER_CLASS_DEBUG  2



COUNTERS_DEF(COUNTERS_ID)
{
  /* these 4 counters must not move */
  COUNTER(lanai_uptime, NORMAL, "Lanai uptime (seconds)")
  COUNTER(counters_uptime, NORMAL, "Counters uptime (seconds)")
  COUNTER(net_send_kbytes, NORMAL, "Net send KBytes")
  COUNTER(net_recv_kbytes, NORMAL, "Net recv KBytes")
  
#if MCP_MYRINET
  COUNTER(net_send_ether_unicast, NORMAL, "Ethernet Unicast send")
  COUNTER(net_send_ether_multicast, NORMAL, "Ethernet Multicast send")
#else
  COUNTER(net_send_ether_native, NORMAL, "Ethernet send")
#endif
  COUNTER(ether_recv_small, NORMAL, "Ethernet Small recv")
  COUNTER(ether_recv_big, NORMAL, "Ethernet Big recv")
  COUNTER(ether_down, NORMAL, "Ethernet recv down")
  COUNTER(ether_overrun, NORMAL, "Ethernet recv overrun")
  COUNTER(ether_rerecv, RAW, "Ethernet re-recv")
  COUNTER(ether_oversized, RAW, "Ethernet recv oversized")


  COUNTER(snf_send, NORMAL, "SNF send pkts")
  COUNTER(snf_recv, NORMAL, "SNF recv pkts")
  COUNTER(snf_drop_prefetch, RAW,  "SNF drop prefetch race")
  COUNTER(snf_drop_flow, NORMAL, "SNF drop ring full")
  COUNTER(snf_drop_desc, RAW, "SNF drop desc credits")
  COUNTER(snf_drop_data, RAW, "SNF drop data credits")



  COUNTER(net_send_raw, RAW, "Net send Raw")
  

  COUNTER(intr, NORMAL, "Interrupts")
  COUNTER(wake_intr, RAW, "Wake interrupt")
  COUNTER(wake_race, RAW, "Wake race")
  COUNTER(wake_closed, RAW, "Wake endpoint closed")
  COUNTER(net_send_queued, RAW, "Net send queued")
  COUNTER(eventq_full, RAW, "Event Queue full")
  COUNTER(rx_dataq_race, RAW, "RX DataQ race")
  COUNTER(fragmented_request, RAW, "Fragmented request")

  COUNTER(net_bad_crc32, NORMAL, "Net bad PHY/CRC32 drop")
#if MCP_MYRINET
  COUNTER(net_bad_crc8, NORMAL, "Net bad CRC8 drop")
  COUNTER(net_mapper_drop, NORMAL, "Net mapper drop")
#endif
  COUNTER(net_overflow_drop, NORMAL, "Net overflow drop")
  COUNTER(net_ether_pause, NORMAL, "Net Recv PAUSEs")
  COUNTER(net_recv_alt, RAW, "Net recv alternate channel")
  COUNTER(net_alt_drop, RAW, "Net Alt drop")

  COUNTER(ether_filter_multicast, NORMAL, "Ethernet Multicast filter drop")
  COUNTER(ether_filter_unicast, NORMAL, "Ethernet Unicast filter drop")

  COUNTER(out_of_send_handles, DEBUG, "Out of send handles")
  COUNTER(ureq_type_unknown, DEBUG, "User request type unknown")
  COUNTER(spurious_request, DEBUG, "Spurious user request")

#if MCP_MYRINET
  COUNTER(net_send_probe, NORMAL, "Net send Probe")
  COUNTER(net_send_probe_ack, NORMAL, "Net send Probe Ack")
  COUNTER(net_send_probe_nack, NORMAL, "Net send Probe Nack")
  COUNTER(net_recv_probe_nack, NORMAL, "Net recv Probe Nack")
  COUNTER(route_dispersion, NORMAL, "Route dispersion")

  COUNTER(route_update, NORMAL, "Route Table update")
#endif

#if MAL_DEBUG
  COUNTER(devel_pkt_loss, DEBUG, "Packet loss (Devel !)")

  COUNTER(foo1, DEBUG, "Foo1")
  COUNTER(foo2, DEBUG, "Foo2")
  COUNTER(foo3, DEBUG, "Foo3")
#endif

};

#endif /* _counters_h_ */
