/*************************************************************************
 * The contents of this file are subject to the MYRICOM MYRINET          *
 * EXPRESS (MX) NETWORKING SOFTWARE AND DOCUMENTATION LICENSE (the       *
 * "License"); User may not use this file except in compliance with the  *
 * License.  The full text of the License can found in LICENSE.TXT       *
 *                                                                       *
 * Software distributed under the License is distributed on an "AS IS"   *
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied.  See  *
 * the License for the specific language governing rights and            *
 * limitations under the License.                                        *
 *                                                                       *
 * Copyright 2003 - 2011 by Myricom, Inc.  All rights reserved.          *
 *************************************************************************/

#ifndef _mx_instance_h_
#define _mx_instance_h_

#include "mal.h"
#include "mal_io.h"
#include "mx_pin.h"
#include "mx_misc.h"
#include "mx_pci.h"
#include "mcp_requests.h"
#include "mcp_wrapper_common.h"

#ifndef MX_DISABLE_ETHERNET
#define MX_DISABLE_ETHERNET 0
#endif
#ifndef MX_HAS_PCIE_LINK_RESET
#define MX_HAS_PCIE_LINK_RESET 0
#endif

/* forward declaration */
typedef struct mx_instance_state mx_instance_state_t;
typedef struct mx_endpt_state mx_endpt_state_t;
typedef struct mx_waitq mx_waitq_t;
typedef struct mx_page_pin mx_page_pin_t;
typedef struct mx_mcp_dma_chunk mx_mcp_dma_chunk_t;
typedef struct mx_host_dma_win mx_host_dma_win_t;
struct mx_ether;

#define MYRI_COPYBLOCK_F_STATIC  0x1
#define MYRI_COPYBLOCK_F_KALLOC  0x2
#define MYRI_COPYBLOCK_F_HUGE    0x4
typedef struct mx_copyblock
{
  void *alloc_addr;
  void *addr;
  uint32_t size;
  int flags;
  mx_page_pin_t *pins;
  mcp_dma_addr_t *dmas;
#if MAL_OS_WINDOWS
  PMDL mdl;
#endif
} mx_copyblock_t;

typedef struct mx_huge_copyblock
{
#if MAL_OS_LINUX || MAL_OS_UDRV
  struct page **pages;
#elif MAL_OS_WINDOWS
  PMDL mdl;
#endif
  void *kva;
  uint64_t size;
  mx_page_pin_t *pins;
  mx_copyblock_t cb;
  unsigned long pinned_pages;
} mx_huge_copyblock_t;

#define mx_cb_to_huge_to_copyblock(cbptr) \
  (struct mx_huge_copyblock *)((uintptr_t)(cbptr) - offsetof(mx_huge_copyblock_t, cb))

typedef struct myri_copyblock
{
  enum { MYRI_CB_K = 0, MYRI_CB_K_HUGE } type;
  int cursize;
  int defsize;
  int flags;
  union {
    mx_huge_copyblock_t hcb;
    mx_copyblock_t cb;
  } u;
} myri_copyblock_t;

/* ops support poll mechanism */
#define MYRI_EOP_F_POLL 0x1

typedef struct myri_endpt_ops {
  /* es->is already set and is lock is held, do necessary to set the type of
   * endpoint being allocated */
  int  flags;
  int  (*alloc)(mx_endpt_state_t *es, mx_instance_state_t *is,
                myri_set_endpt_t *set_endpt);
  int  (*enable)(mx_endpt_state_t *es, uaddr_t arg);
  int  (*ioctl)(mx_endpt_state_t *es, uint32_t cmd, uaddr_t in);
  int  (*mmap)(mx_endpt_state_t *es, unsigned long req, unsigned long *off,
               void **kva, int *mem_type, int *readonly, mx_page_pin_t **pin);
  int  (*poll_events_pending)(mx_endpt_state_t *es);
  void (*poll_prep_wait)(mx_endpt_state_t *es);
  int  (*poll_wait)(mx_endpt_state_t *es, int timeout_ms, int events);
  void (*interrupt_is)(mx_instance_state_t *is, void *context);
  int  (*interrupt_ep)(mx_endpt_state_t *es, int type);
  void (*close)(mx_endpt_state_t *es);
} myri_endpt_ops_t;

typedef struct mx_intr
{
  uint32_t slot;
  uint16_t seqnum;
  uint32_t spurious;
  uint32_t cnt;
  mcp_intr_t *ring;
  mx_page_pin_t pin;
} mx_intr_t;

#include "myri_snf_io.h"
#include "myri_snf_fwd.h"

typedef struct mx_routes
{
  char *host_table;
  char *mcp_table;
  int *indexes;
  int *offsets;
  int block_size;
} mx_routes_t;

typedef int (mx_init_board_t)(mx_instance_state_t *is);
typedef int (mx_map_board_t)(mx_instance_state_t *is);
typedef void (mx_unmap_board_t)(mx_instance_state_t *is);
typedef void (mx_claim_interrupt_t)(mx_instance_state_t *is, uint8_t endpt_irq, uint8_t endpt);
typedef int (mx_read_irq_level_t)(mx_instance_state_t *is);
typedef void (mx_enable_interrupt_t)(mx_instance_state_t *is);
typedef void (mx_disable_interrupt_t)(mx_instance_state_t *is);
typedef void (mx_park_board_t)(mx_instance_state_t *is);
typedef int (mx_detect_parity_t)(mx_instance_state_t *is);
typedef void (mx_get_freq_t)(mx_instance_state_t *is);
typedef void (mx_write_kreq_t)(mx_instance_state_t *is, mcp_kreq_t *kreq);
typedef uint32_t (mx_get_rtc_t)(mx_instance_state_t *is);

typedef struct mx_board_ops {
  mx_init_board_t 		*init;
  mx_map_board_t 		*map;
  mx_unmap_board_t 		*unmap;
  mx_claim_interrupt_t 		*claim_interrupt;
  mx_read_irq_level_t		*read_irq_level;
  mx_enable_interrupt_t 	*enable_interrupt;
  mx_disable_interrupt_t 	*disable_interrupt;
  mx_park_board_t		*park;
  mx_detect_parity_t		*detect_parity_error;
  mx_get_freq_t		        *get_freq;
  mx_write_kreq_t		*write_kreq;
  mx_get_rtc_t			*get_rtc;
} mx_board_ops_t;

extern mx_board_ops_t mx_lx_ops;
extern mx_board_ops_t mx_lz_ops;
void mx_lx_init_board_ops(void);
void mx_lz_init_board_ops(void);
void mx_lz_read_mac(mx_instance_state_t *is, char mac[18]);
int mx_maccmp(const char *mac_a, const char *list_of_mac_b_substr);
void mx_lz_phys_id(mx_instance_state_t *is, int id_on);

#define MX_ADDRS_PER_VPAGE  (MX_VPAGE_SIZE / sizeof(mcp_dma_addr_t))
#define MX_ADDRS_PER_VPAGE_SHIFT  (MX_VPAGE_SHIFT - MX_DMA_ADDR_SHIFT)

#define MX_NUM_DIRECT_DMA  4  /* XXXX move me elsewhere!!!! */
#define MX_MAX_MAPPER_TYPES 5


/* An mx_dma_table contains one page of DMA addresses.
   Lengths for each DMA are implicitly assumed to be VPAGE size.
   The MCP accesses this structure, so it must be contiguous in DMA space.  
   The table can be a pte table or a higher-level table */

struct mx_dma_table
{
  mx_page_pin_t pin;
  mcp_dma_addr_t *desc;		/* pointer to DMA page */
  union {
    mx_page_pin_t *pins; /* DMA pin info for data pages */
    struct mx_dma_table **tables; /* tables of inferor lvel  */
  } u;
  int nb_entries;
  int nb_subtables;
  int log_level; /* 0=> table of data pages , n*MX_ADDRS_PER_VPAGE_SHIFT => n-higher-level table */
};

typedef struct mx_dma_table mx_dma_table_t;

/* An mx_host_dma_win_t is a container for one rdma window,
   it contains the page table associated with this window. */

struct mx_host_dma_win
{
  uint32_t flags;
  uint32_t length;
  uint32_t seqnum;
  struct mx_dma_table table;
};

typedef struct myri_raw_tx_buf
{
  STAILQ_ENTRY(myri_raw_tx_buf) next;
  mcp_dma_addr_t dma;
  void *buf;
  uint32_t id;
  uint64_t context;
} myri_raw_tx_buf_t;

STAILQ_HEAD(myri_raw_tx_bufq, myri_raw_tx_buf);

/* HW clock functionality */
int  myri_clock_init(mx_instance_state_t *is);
void myri_clock_destroy(mx_instance_state_t *is);
int  myri_clock_sync_now(mx_instance_state_t *is);
void myri_clock_intr(mx_instance_state_t *is);

int myri_secs_to_year(uint32_t seconds);
uint32_t myri_year_to_secs(int year);

#if MYRI_ENABLE_PTP
enum myri_ptp_state {
  MYRI_PTP_STOPPED  = -1,
  MYRI_PTP_STARTING = 0,
  MYRI_PTP_RUNNING  = 1
};
#endif

#define MX_IS_DEAD        (1 << 0)
#define MX_IS_LOGGING     (1 << 1)
#define MX_HOSTNAME_SET   (1 << 2)
#define MX_PARITY_RECOVERY (1 << 3)

#define MYRI_WAIT_RX_EVENT  1
#define MYRI_WAIT_TX_EVENT  2

struct ep_irq {
  mx_intr_t intr;
  struct mx_instance_state *is;
  mx_endpt_state_t **ep;
};

struct mx_instance_state
{
  unsigned int id;            /* unit number.         */
  unsigned int flags;         /* Flags, defined above */
  mx_board_ops_t board_ops;

  /*************
   * LANai access fields
   *************/

  /* Record of lanai mappings */
    
  struct
  {
    /* This pointer is opaque to ensure that the special registers
       are referenced only through the pio.h interface. */
    void *special_regs;
    volatile uint32_t *control_regs;
    volatile uint8_t *sram;
#if MX_DMA_DBG
    volatile uint32_t *dma_dbg_bitmap;
    volatile uint32_t dma_dbg_lowest_addr;
#endif
    /* The globals are not a structure to insure that they are
       referenced only throught the pio.h interface. */
    volatile void *globals; /* Set iff MCP running. */
    uint32_t running;
    char *eeprom_strings;
    char *product_code;
    char *part_number;
    char *serial;
  } lanai;

  /* Pointer to LANai interface card, mapped into kernel memory. */

  uint32_t board_span;
  uint32_t sram_size;
  uint32_t specials_size;
  uint32_t using_msi:1;
  uint32_t msi_enabled:1;
  uint32_t msix_enabled:1;
  uint32_t cpu_freq;
  uint32_t pci_freq;
  mx_arch_instance_info_t arch; /* Architecture-specific field. */
  mx_sync_t sync;
  mx_endpt_state_t **es;
  int es_count;
  char is_name_default[8];
  const char *is_name;
  uint8_t mac_addr[8];
  char *mac_addr_string;

#ifdef MYRI_MSI
  struct {
	char *name;
	void *vectors;
  } msix;
#endif

  mx_intr_t intr;
  struct ep_irq *ep_irq;

  struct {
    mcp_kreq_t kreq;
    mx_spinlock_t spinlock;
  } cmd;
  mcp_kreq_t *kreqq;
  int kreqq_submitted;
  int kreqq_max_index;
  mx_spinlock_t kreqq_spinlock;
  struct {
    mx_sync_t sync;
    uint32_t size;
  } logging;
  mx_sync_t init_sync;
  mx_atomic_t ref_count;
  struct {
    mx_endpt_state_t *es;
    int use_count;
    mx_spinlock_t spinlock;
    mx_sync_t sync;
    int wakeup_needed;
    myri_raw_tx_buf_t *tx_bufs;
    struct myri_raw_tx_bufq tx_bufq;
    mx_copyblock_t tx_cb;
    uint32_t tx_req;
    uint32_t tx_done;
    mx_copyblock_t rx_cb;
    uint32_t rx_cnt;
    mcp_raw_desc_t *descs;
    mx_page_pin_t descs_pin;
    uint32_t desc_cnt;
  } raw;
  uint32_t num_ports;
  uint8_t pci_rev;
  uint16_t pci_devid;
  uint32_t bar0_low;
  uint32_t bar0_high;
  uint32_t bar2_low;
  uint32_t bar2_high;
  uint32_t msi_addr_low;
  uint32_t msi_addr_high;
  uint16_t msi_data;
  uint16_t msi_ctrl;
  uint16_t exp_devctl;
  uint32_t ze_print_pos;
  int board_type;
  uint32_t mcp_print_len;
  uint32_t mcp_print_addr;
  uint32_t mcp_print_idx;
  char *mcp_print_buffer;
  uint32_t old_lanai_uptime;
  struct mx_ether *ether;
  int ether_is_open;
  struct {
    mx_sync_t wait_sync;
    uint32_t cycles;
    uint32_t count;
    int busy;
  } dmabench;
  int parity_errors_detected;
  int parity_errors_corrected;
  uintptr_t parity_critical_start;
  uintptr_t parity_critical_end;
  mx_sync_t ready_for_recovery_sync;
  struct {
    int command_tweaked;
    uint16_t command_offset;
    uint16_t command_orig_val;
  } pci_cap;
  uint32_t link_state;
  struct {
    int reason;
    int arg;
  } saved_state;
  struct {
    mx_page_pin_t pin;
    char *addr;
  } tx_done;
  struct {
    mx_page_pin_t pin;
    char *addr;
  } sysbox;
  struct  {
    uint32_t seqnum;
    mx_spinlock_t spinlock; /* For accumulation */
    mx_sync_t sync;
    mx_page_pin_t pin;
    char *addr;      /* host DMA copy */
    uint32_t *mcp;   /* SRAM access (pio read) */
    uint64_t *accum; /* Accumulated counts in driver */
    uint32_t *accum_last;  /* Keep track of last value read */
    uint32_t cnt;
  } counters;
  struct {
    mx_page_pin_t pin;
    char *addr;
  } bogus;
  mx_kernel_window_t *kernel_window;
  uint32_t *dma_dbg_bitmap_host;
  int manual_dispersion;
  uint32_t crc32_table[16];

  myri_snf_stats_t   snf_stats;
  myri_snf_stats_t   snf_stats_prev;
  myri_snf_rx_state_t   snf;
  void              *snf_rx_doorbell;


  /* NIC and host clock synchronization */
  mx_spinlock_t           clk_spinlock;
  enum myri_timesource_state clk_ts_state;
  uint32_t                clk_db_offset;
  int                     clk_hwclock_seqno;
  int                     clk_sync_init_cnt;
  uint64_t                clk_sync_init_nsecs;
  mcp_hwclock_update_t   *clk_update;
  myri_clksync_params_t   clkp_latest;
#if MYRI_ENABLE_PTP
  enum myri_ptp_state     ptp_state;
  mx_spinlock_t           ptp_state_lock;
  struct myri_ptp        *ptp;
#endif

  /* enable/disable use of timesource */
  int                     timesource_enable;
  
  /* data used in the synching of the timesource */
  struct {
    uint64_t nsecs;
    uint64_t cycles;
    int year;
    int attempts;
  } timesource_data;

#if MAL_OS_LINUX && defined MYRI_PPS_SUPPORT
  /* pps stuff */
  MYRI_PPS_TYPE pps_dev;
  uint64_t pps_last_second;
  int pps_timesource_good;
  struct timer_list pps_timer;
#endif /* MAL_OS_LINUX && MYRI_PPS_SUPPORT */
};

#define MX_RDMA_WIN_PER_PAGE (MX_VPAGE_SIZE/sizeof(mcp_rdma_win_t))

#define MX_ES_SOFT      (1 << 0)
#define MX_ES_CLOSING	(1 << 1)
#define MX_ES_RECOVERY	(1 << 2)
#define MX_ES_MCP_SESSION (1 << 3)
#define MX_ES_MCP_ENABLED (1 << 4)
#define MX_ES_EARLY_ENABLE (1 << 5) /* XXX: will be removed */

struct mx_endpt_state
{
  mx_instance_state_t *is;
  enum myri_endpt_type es_type;
  uint32_t endpt;
  uint32_t flags;
  uint32_t ref_count;  /* access protected by sync */
  mx_sync_t sync;
  myri_endpt_ops_t *eops;

  mx_sync_t wait_sync;
  mx_sync_t close_sync;
  mx_sync_t vq_sync;  /* prehash */

  myri_endpt_flow_t *flow_window;
  mx_page_pin_t flow_window_pin;	

  /* MD handle */
  mx_arch_endpt_info_t arch;

  /* copyblocks */
  mx_copyblock_t user_mmapped_sram;  /* sram, not kernel mem */
  mx_copyblock_t user_mmapped_zreq;  /* mmio, not kernel mem */
  mx_copyblock_t sendq;
  mx_copyblock_t eventq;
  mx_huge_copyblock_t recvq;

  uint32_t privileged;
  uint32_t is_kernel;
#ifdef _LINUX_UIDGID_H
  kuid_t euid;
#else
  uint32_t euid;
#endif
  myri_endpt_opener_t opener;
  int endpt_error;
  uint32_t session_id;

  /* If endpoint is pinned to a cpu, keep track of its cpuid and nodeid, if
   * applicable */
  int cpu_id;
  int node_id;

  union snf_endpt_state snf;
};

int mx_init_driver(void);
int mx_finalize_driver(void);

int  mx_instance_init (mx_instance_state_t *is, int32_t unit);
int  mx_instance_finalize (mx_instance_state_t *is);
void mx_lanai_print(mx_instance_state_t *is, int idx);
int mx_load_mcp(mx_instance_state_t *is, void * img, int limit, int *actual_len);


void *mx_map_pci_space (mx_instance_state_t * is, int bar, uint32_t offset, 
		       uint32_t len);

void *mx_map_io_space (mx_instance_state_t * is, uint32_t offset, 
		       uint32_t len);
void  mx_unmap_io_space (mx_instance_state_t * is,
			 uint32_t len, void *kaddr);

void mx_dump_interrupt_ring(mx_instance_state_t *is);


extern mx_instance_state_t **mx_instances;
extern uint32_t myri_max_instance;
extern uint32_t mx_num_instances;
extern uint32_t mx_mapper_mask;
extern uint32_t mx_mapper_running_map;
extern uint32_t mx_mapper_level_mask;
extern uint32_t mx_mapper_verbose;
extern int mx_mapper_msgbuf_size;
extern int mx_mapper_mapbuf_size;
extern uint32_t myri_mx_small_message_threshold;
extern uint32_t myri_mx_medium_message_threshold;
extern uint32_t myri_security_disabled;
extern int mx_kernel_mapper;
extern uint32_t myri_z_loopback;
extern uint32_t myri_ethernet_bitmap;
extern uint32_t myri_serdes_mode;
extern int mx_has_xm;
extern char * myri_mac;
extern uint32_t myri_pbkey;

#if MAL_OS_MACOSX || MAL_OS_SOLARIS
#define MX_MAX_INSTANCE_DEFAULT 4
#else
#define MX_MAX_INSTANCE_DEFAULT 32
#endif

extern uint32_t myri_mx_max_nodes;
extern uint32_t myri_max_endpoints;
extern uint32_t myri_recvq_vpage_cnt;
extern uint32_t myri_eventq_vpage_cnt;

#define MX_MAX_ENDPOINTS_DEFAULT 16

extern uint32_t myri_intr_coal_delay;

extern mx_atomic_t myri_max_user_pinned_pages;
extern uint32_t myri_mx_max_send_handles;
extern uint32_t mx_max_pull_handles;
extern uint32_t mx_max_push_handles;
extern uint32_t myri_mx_max_rdma_windows;
extern uint32_t mx_cacheline_size;
extern uint32_t myri_override_e_to_f;
extern uint32_t myri_parity_recovery;
extern uint32_t myri_recover_from_all_errors;
extern uint32_t myri_pcie_down;
extern uint32_t myri_pcie_down_on_error;
extern uint32_t myri_mx_max_host_queries;
extern uint32_t myri_dma_pfn_max;
extern uint32_t myri_force;
extern uint32_t myri_jtag;
extern uint32_t myri_mx_max_macs;
extern uint32_t myri_irq_sync;
extern uint32_t myri_msi_level;
extern uint32_t myri_ether_pause;
extern uint32_t myri_idt_war;
extern uint32_t myri_clksync_period;
extern uint32_t myri_clksync_error;
extern uint32_t myri_clksync_error_count;
extern uint32_t myri_clksync_verbose;
extern uint32_t myri_timesource_loss_mode;
extern uint32_t myri_snf_rings;
extern uint32_t myri_snf_flags;

extern mx_spinlock_t mx_lanai_print_spinlock;

extern char mx_default_hostname[MYRI_MAX_STR_LEN];

/* sleeping and waking up */

#define MX_SLEEP_INTR    1
#define MX_SLEEP_NOINTR  2

/* for the watchdog to properly detect parity error (rather than such
   error ending up as some MX_DEAD_xxx_TIMEOUT), the final timeout
   should be:
   final-timeout >= (min-desired-timeout + 2*Watchdog)
   in addition to sram parity error, this avoid confusing root cause
   and secondary effects for any HW check done in the watchdog_body
*/

#define MX_SMALL_WAIT    100
#define MX_CLOSE_WAIT	 1000*30 /* 30 seconds */
#define MX_LOGGING_WAIT  MX_MAX_WAIT

#if MAL_OS_UDRV
#define MCP_INIT_TIMEOUT	(mx_lxgdb ? 600*1000 : 12*1000)
#define MCP_COMMAND_TIMEOUT	(mx_lxgdb ? 600*1000 : 10)
#define MX_WATCHDOG_TIMEOUT	(mx_lxgdb ? 10 : 5)
#define MX_LXGDB		mx_lxgdb
#else
/* FIXME: do we need 10s timeout ? */
#define MCP_INIT_TIMEOUT	12*1000
#define MCP_COMMAND_TIMEOUT	10
#define MX_WATCHDOG_TIMEOUT	5
#define MX_LXGDB		0
#endif


void mx_wake(mx_sync_t *);
int mx_wake_once(mx_sync_t *); /* returns 1 if notify happened */
int mx_sleep(mx_sync_t *, int, int flags);
mx_waitq_t *mx_new_waitq(void);
void mx_free_waitq(mx_waitq_t *);

/* syncs.. need to move elsewhere */

void mx_sync_init (mx_sync_t *s, mx_instance_state_t *is, int unique, 
		   char *string);
void mx_sync_reset (mx_sync_t *s);
void mx_sync_destroy (mx_sync_t *s);

#ifndef mx_spin_lock_destroy
void mx_spin_lock_destroy(mx_spinlock_t *s);
#endif

void mx_spin_lock_init(mx_spinlock_t *s, mx_instance_state_t *is, int unique, 
		       char *string);

#ifndef mx_mutex_enter
void mx_mutex_enter (mx_sync_t *s);
int  mx_mutex_try_enter (mx_sync_t *s); /* returns 0 if successful */
void mx_mutex_exit (mx_sync_t *s);
#endif

extern mx_sync_t mx_global_mutex;

/* open/close/ioctl .. need to move elsewhere */


int mx_common_ioctl(mx_endpt_state_t *es, uint32_t cmd, const uaddr_t in);

int mx_endptless_ioctl(uint32_t cmd, const uaddr_t in,
		       uint32_t privileged, uint32_t is_kernel);

int myri_common_open_args(int cmd, uaddr_t arg, myri_set_endpt_t *);

int myri_common_open(int32_t unit, mx_endpt_state_t *es, 
                     myri_set_endpt_t *set_endpt, uaddr_t arg);

void mx_common_close (mx_endpt_state_t *es);

void mx_unpin_page(mx_instance_state_t *, mx_page_pin_t *, int);
void mx_unpin_vpages(mx_instance_state_t *, mx_page_pin_t *, int, int);
int mx_pin_page(mx_instance_state_t *, mx_page_pin_t *, int, uint64_t);
int mx_pin_vpages(mx_instance_state_t *, mx_page_pin_t *, mcp_dma_addr_t *, int, int, uint64_t);
mx_page_pin_t *mx_find_pin(mx_endpt_state_t *es,  
			   int (*testfp)(mx_page_pin_t *pin, void *arg), void *arg);

/* PCI accessor functions */
int mx_read_pci_config_8 (mx_instance_state_t *is,
			  uint32_t offset,
			  uint8_t *val);

int mx_read_pci_config_16 (mx_instance_state_t *is,
			   uint32_t offset,
			   uint16_t *val);

int mx_read_pci_config_32 (mx_instance_state_t *is,
			   uint32_t offset,
			   uint32_t *val);

int mx_write_pci_config_8 (mx_instance_state_t *is,
			   uint32_t offset,
			   uint8_t val);

int mx_write_pci_config_16 (mx_instance_state_t *is,
			    uint32_t offset,
			    uint16_t val);

int mx_write_pci_config_32 (mx_instance_state_t *is,
			    uint32_t offset,
			    uint32_t val);


#if MAL_OS_UDRV
void mx_write_lanai_isr(struct mx_instance_state *is, uint32_t val);
void mx_write_lanai_ctrl_bit(struct mx_instance_state *is, int val);
#else
#define mx_write_lanai_isr(is,val) mx_write_lanai_special_reg_u32(is,lx.ISR, val)
#define mx_write_lanai_ctrl_bit(is,bit) mx_write_pci_config_32(is, 0x40, 1U << (bit))
/* or equivalent: is->lanai.control_regs[0x10] = htonl(1 << MX_LX_CTRL(bit)))) */
#endif

#define MX_MEM_HOSTMEM 0
#define MX_MEM_SRAM 1
#define MX_MEM_CONTROL 2
#define MX_MEM_SPECIAL 3
#define MX_MEM_HOSTMEM_UNMAPPED 4

int mx_mmap_off_to_kva(mx_endpt_state_t *, unsigned long, void **, int *, int *,
		       mx_page_pin_t **);

int mx_mcp_command(mx_instance_state_t *, uint8_t, uint32_t, uint32_t, 
		   uint32_t, uint32_t *);
void mx_mcp_command_done(mx_instance_state_t *, mcp_cmd_status_t, uint32_t);
  
int mx_common_interrupt(mx_instance_state_t *is);

#ifdef MYRI_MSI
int mx_common_msix_ep_irq(struct ep_irq *emi);
#endif

/* memory registration related functionality */
void mx_free_dma_win (mx_instance_state_t *is, mx_host_dma_win_t *hc);
mx_host_dma_win_t *mx_new_dma_win(mx_instance_state_t *is, int nsegs);

static inline int
mx_is_dead(mx_instance_state_t *is)
{
  return (is->flags & (MX_IS_DEAD | MX_PARITY_RECOVERY));
}
#define mx_digit(c) (((c) >= '0' && (c) <= '9') ? ((c) - '0') : \
                 (((c) >= 'A' && (c) <= 'F') ? (10 + (c) - 'A') : \
                  (((c) >= 'a' && (c) <= 'f') ? (10 + (c) - 'a') : -1)))

int mx_alloc_dma_page(mx_instance_state_t *is, char **addr, mx_page_pin_t *pin);

void mx_free_dma_page(mx_instance_state_t *is, char **addr, mx_page_pin_t *pin);
int mx_alloc_zeroed_dma_page(mx_instance_state_t *is, char **addr, mx_page_pin_t *pin);
#if MAL_OS_LINUX
int mx_alloc_zeroed_dma_page_node(mx_instance_state_t *is, char **addr, mx_page_pin_t *pin, int node_id);
#else
#define mx_alloc_zeroed_dma_page_node mx_alloc_zeroed_dma_page
#endif
int mx_alloc_zeroed_dma_page_relax_order(mx_instance_state_t *is, char **addr, mx_page_pin_t *pin);

#ifdef MX_HAS_RELAX_ORDER_DMA
int mx_alloc_dma_page_relax_order(mx_instance_state_t *is, char **addr, mx_page_pin_t *pin);
#else
#define mx_alloc_dma_page_relax_order mx_alloc_dma_page
#define mx_alloc_copyblock_relax_order mx_alloc_copyblock
#endif

int mx_rand(void);
void mx_watchdog_body(void);


int mx_dma_map_copyblock(mx_endpt_state_t *es, mx_copyblock_t *cb, uint32_t offset);

int mx_common_alloc_copyblock(mx_instance_state_t *is, mx_copyblock_t *cb);
void mx_common_free_copyblock(mx_instance_state_t *is, mx_copyblock_t *cb);
void mx_parse_mcp_error(mx_instance_state_t *is);
#if MAL_OS_LINUX
void *mx_map_huge_copyblock(mx_huge_copyblock_t *hcb);
void mx_unmap_huge_copyblock(mx_huge_copyblock_t *hcb);
#else
#define mx_map_huge_copyblock(hcb) NULL
#define mx_unmap_huge_copyblock(hcb) do { } while (0)
#endif


/* copyin and out */
static inline int
mx_copyin(uaddr_t what, void *where, size_t amount, uint32_t is_kernel)
{
  if (is_kernel) {
    bcopy((void *)(uintptr_t)what, where, amount);
    return 0;
  } else
    return mx_arch_copyin(what, where, amount);
}

static inline int
mx_copyout(void *what, uaddr_t where, size_t amount, uint32_t is_kernel)
{
  if (is_kernel) {
    bcopy(what, (void *)(uintptr_t)where, amount);
    return 0;
  } else
    return mx_arch_copyout(what, where, amount);
}

int mx_instance_status_string(int unit, char **str, int *len);

void mx_set_default_hostname(void);

#define mx_start_mapper(is)
#define mx_stop_mapper(is)

#if MAL_OS_LINUX
void myri_update_numa_config(mx_endpt_state_t *es);
#else
static inline void myri_update_numa_config(mx_endpt_state_t *es)
{
  es->cpu_id = -1;
  es->node_id = -1;
}
#endif


mx_endpt_state_t *mx_get_endpoint(mx_instance_state_t *is, int endpt);
void mx_put_endpoint(mx_endpt_state_t *es);
int mx_direct_get(mx_endpt_state_t *dst_es, mx_shm_seg_t *dst_segs, uint32_t dst_nsegs,
		  mx_endpt_state_t *src_es, mx_shm_seg_t *src_segs, uint32_t src_nsegs,
		  uint32_t length);
mx_instance_state_t *mx_get_instance(uint32_t unit);
void mx_release_instance(mx_instance_state_t *is);
void mx_mark_board_dead(mx_instance_state_t *is, int reason, int arg);
void mx_parse_eeprom_strings(mx_instance_state_t *is);

int mx_direct_get_common(mx_shm_seg_t *dst_segs, uint32_t dst_nsegs,
			 void * src_space, mx_shm_seg_t *src_segs, uint32_t src_nsegs,
			 uint32_t length);

/* OS specific callback for direct get, copying from another process
 * user-space to current process user-space.
 */
int mx_arch_copy_user_to_user(uaddr_t udst, uaddr_t usrc, void * src_space,
			      uint32_t len);


int mx_pcie_link_reset(mx_instance_state_t *is);
int mx_intr_recover(mx_instance_state_t *is);
uint16_t mx_bridge_pci_sec_status(mx_instance_state_t *is);
uint32_t mx_bridge_id(mx_instance_state_t *is);
int mx_find_capability(mx_instance_state_t *is, unsigned cap_id);

int myri_select_mcp(int board_type, const unsigned char **mcp,
		    int *len, unsigned int *parity_critical_start,
		    unsigned int *parity_critical_end);
int myri_counters(int board_type, const char ***counters, uint32_t *count,
                  const uint32_t **cclass);

void myri_counters64_update(mx_instance_state_t *is);
void myri_counters64_clear(mx_instance_state_t *is);
int  myri_counters64_get(mx_instance_state_t *is, uint64_t *c_o, int num_counters,
                         enum myri_counters_get_type type);

int myri_check_license(mx_instance_state_t *is, uint32_t *valid);

int
mx_alloc_huge_copyblock(mx_instance_state_t *is, mx_huge_copyblock_t *hcb,
			unsigned long lib_offset, int node_id);

void
mx_free_huge_copyblock(mx_instance_state_t *is, mx_huge_copyblock_t *hcb);

void myri_snf_get_stats(mx_instance_state_t *is, myri_snf_stats_t *stats_o);
void myri_snf_clear_stats(mx_instance_state_t *is, int reset);


int myri_atoi(char *s);
int myri_endpt_get_common_copyblocks(const mx_endpt_state_t *es, mx_get_copyblock_t *desc,
                                     uint32_t *offp);
int myri_endpt_enable_common_copyblocks(mx_endpt_state_t *es);
int myri_endpt_enable_mcp(mx_endpt_state_t *es);

#if MX_KERNEL_LIB
#define mx_endpt_alloc_copyblock(es,cb)					\
  ((es)->is_kernel ? mx_common_alloc_copyblock((es)->is,cb)		\
   : mx_alloc_copyblock((es)->is,cb))
#define mx_endpt_alloc_copyblock_relax_order(es,cb)			\
  ((es)->is_kernel ? mx_common_alloc_copyblock((es)->is,cb)		\
   : mx_alloc_copyblock_relax_order((es)->is,cb))
#define mx_endpt_free_copyblock(es,cb)			   \
  ((es)->is_kernel ? mx_common_free_copyblock((es)->is,cb) \
   : mx_free_copyblock((es)->is,cb))
#define mx_endpt_alloc_huge_copyblock(es,cb)				\
   mx_alloc_huge_copyblock((es)->is,cb,0,-1)
#define mx_endpt_alloc_huge_copyblock_relax_order(es,cb)		\
   mx_alloc_huge_copyblock_relax_order((es)->is,cb,0)
#define mx_endpt_free_huge_copyblock(es,cb)			\
   mx_free_huge_copyblock((es)->is,cb)
#else
#define mx_endpt_alloc_copyblock(es,cb) mx_alloc_copyblock((es)->is,cb)
#define mx_endpt_alloc_copyblock_relax_order(es,cb) mx_alloc_copyblock_relax_order((es)->is,cb)
#define mx_endpt_free_copyblock(es,cb) mx_free_copyblock((es)->is,cb)
#if MAL_OS_WINDOWS
int mx_win_alloc_huge_copyblock(mx_instance_state_t *is, 
				mx_huge_copyblock_t *hcb,
				unsigned long lib_offset,
				int node_id);
void mx_win_free_huge_copyblock(mx_instance_state_t *is,
				mx_huge_copyblock_t *hcb);
#define mx_endpt_alloc_huge_copyblock(es,cb) mx_win_alloc_huge_copyblock((es)->is,cb,0,0)
#define mx_endpt_alloc_huge_copyblock_relax_order(es,cb) mx_win_alloc_huge_copyblock_relax_order((es)->is,cb,0,0)
#define mx_endpt_free_huge_copyblock(es,cb) mx_win_free_huge_copyblock((es)->is,cb)
#else
#define mx_endpt_alloc_huge_copyblock(es,cb) mx_alloc_huge_copyblock((es)->is,cb,0,-1)
#define mx_endpt_alloc_huge_copyblock_relax_order(es,cb) mx_alloc_huge_copyblock_relax_order((es)->is,cb,0)
#define mx_endpt_free_huge_copyblock(es,cb) mx_free_huge_copyblock((es)->is,cb)
#endif
#endif


#endif /* _mx_instance_h_ */
