/*************************************************************************
 * The contents of this file are subject to the MYRICOM MYRINET          *
 * EXPRESS (MX) NETWORKING SOFTWARE AND DOCUMENTATION LICENSE (the       *
 * "License"); User may not use this file except in compliance with the  *
 * License.  The full text of the License can found in LICENSE.TXT       *
 *                                                                       *
 * Software distributed under the License is distributed on an "AS IS"   *
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied.  See  *
 * the License for the specific language governing rights and            *
 * limitations under the License.                                        *
 *                                                                       *
 * Copyright 2003 - 2004 by Myricom, Inc.  All rights reserved.          *
 *************************************************************************/

/* This MX kernel lib code was originally contributed by
 * Brice.Goglin@ens-lyon.org (LIP/INRIA/ENS-Lyon) */

static const char __idstring[] = "@(#)$Id$";

#include "mal_auto_config.h"

#include "mx_arch_klib.h"
#include "mx_arch.h"
#include "mx_instance.h"
#include "mx_malloc.h"
#include "myri_klib.h"

/* all these seem useless for now ? */
int myri_klib_initialized = 0;

#include "snf.h"

int
myri_init_klib(void)
{
  if (!myri_klib_initialized) {
    myri_klib_initialized = 1;
  }
  return 0;
}

void
myri_finalize_klib(void)
{
  if (myri_klib_initialized) {
    myri_klib_initialized = 0;
  }
}

int
myri_klib_open(mx_endpt_state_t **es)
{
  *es = NULL;
  return 0;
}

/* 
 * An NULL es means that the external module only open
 * a board but not endpoint.
 * This leads to mx_endptless_ioctl.
 */
int
myri_klib_set_endpoint(mx_endpt_state_t **retes, uint32_t unit,
                       myri_set_endpt_t *set_endpt, void *context)
{
  int status;
  mx_endpt_state_t *es;

  es = mx_kmalloc(sizeof(*es), MX_MZERO|MX_WAITOK);
  if (es == 0)
    return ENOMEM;

  es->is_kernel = 1;
  /* set to -1.  Can't set to a real pid, as the process opening
     the endpoint could disappear */
  es->opener.pid = -1;

  status = myri_common_open(unit, es, set_endpt, (uaddr_t) context);
  if (status != 0) {
    mx_kfree(es);
    return (status);
  }

  myri_update_numa_config(es);

  *retes = es;

  MAL_DEBUG_PRINT (MAL_DEBUG_OPENCLOSE, 
		  ("Board %d, endpoint %d opened\n", 
		   unit, es->endpt));
  return 0;
}

int
myri_klib_close(mx_endpt_state_t *es)
{
  if (es != NULL ) {
    mx_common_close(es);
    mx_kfree(es);
  }
  return 0;
}

int
myri_klib_ioctl(mx_endpt_state_t *es, int cmd, void *p, size_t size)
{
  int retval;
  if ( es == NULL ) {
    retval = mx_endptless_ioctl(cmd, (uaddr_t) p, 0, 1);
  } else {
    retval = mx_common_ioctl(es, (uint32_t) cmd, (uaddr_t) p);
    if (retval == ENOTTY) {
      retval = mx_endptless_ioctl((uint32_t) cmd, (uaddr_t) p, 0, 1);
    }
  }
  return retval;
}

int
myri_klib_map(mx_endpt_state_t *es, uintptr_t offset, void **ptr, mx_page_pin_t **pin)
{
  int mem_type, readonly, rc;
  void *kva;
  mx_page_pin_t *pin_req;

  if (pin == NULL)
    pin = &pin_req;

  if (offset == MYRI_OFFSET_UNMAPPED)
    return 0;

  if ((rc = mx_mmap_off_to_kva(es, (unsigned long) offset, &kva,
                               &mem_type, &readonly, pin)))
    return rc;
  else {
    *ptr = (void *) kva;
    return 0;
  }
}
